import type { ExceptionMessage } from "#app/exceptions/handler.ts"
import Page, { type PageRaw } from "#app/models/page.ts"
import { throttle } from "#start/limiter.ts"
import type { HttpContext } from "@adonisjs/core/http"
import router from "@adonisjs/core/services/router"
import vine from "@vinejs/vine"

export const get_shortest_paths_validator = vine.compile(
  vine.object({
    fromPageId: vine.number().withoutDecimals().positive(),
    toPageId: vine.number().withoutDecimals().positive(),
  }),
)

interface get_shortest_paths_response {
  /**
   * Object to get page information by their id.
   * - Key: Page id.
   * - Value: Page information.
   */
  pages: Record<number, PageRaw>

  /**
   * Paths between two pages using only internal links.
   * Each path is an array of page ids.
   */
  paths: number[]
}

const getPathsBetweenPages = async (
  fromPage: Page,
  toPage: Page,
): Promise<number[]> => {
  if (fromPage.id === toPage.id) {
    return [fromPage.id]
  }

  // A queue of paths from the start page
  const forwardQueue: Array<[Page, number[]]> = [[fromPage, [fromPage.id]]]
  // A queue of paths from the end page
  const backwardQueue: Array<[Page, number[]]> = [[toPage, [toPage.id]]]

  // Sets to track visited pages from both ends
  const visitedFromStart = new Set<number>([fromPage.id])
  const visitedFromEnd = new Set<number>([toPage.id])

  // Maps to track paths
  const forwardPaths = new Map<number, number[]>([[fromPage.id, [fromPage.id]]])
  const backwardPaths = new Map<number, number[]>([[toPage.id, [toPage.id]]])

  // Helper function to process a queue in one direction
  const processQueue = async (
    queue: Array<[Page, number[]]>,
    visitedThisSide: Set<number>,
    visitedOtherSide: Set<number>,
    pathsThisSide: Map<number, number[]>,
    pathsOtherSide: Map<number, number[]>,
  ): Promise<number[]> => {
    const [currentPage, currentPath] = queue.shift() as [Page, number[]]
    await currentPage.load("internalLinks")

    for (const link of currentPage.internalLinks) {
      if (!visitedThisSide.has(link.id)) {
        const newPath = [...currentPath, link.id]
        visitedThisSide.add(link.id)
        pathsThisSide.set(link.id, newPath)

        // If the other side has visited this page, we've found a meeting point
        if (visitedOtherSide.has(link.id)) {
          const pathFromOtherSide = pathsOtherSide.get(link.id) ?? []
          return [...newPath.slice(0, -1), ...pathFromOtherSide.reverse()]
        }

        // Otherwise, continue the BFS
        queue.push([link, newPath])
      }
    }

    return []
  }

  while (forwardQueue.length > 0 && backwardQueue.length > 0) {
    // Expand the BFS from the start side
    let result = await processQueue(
      forwardQueue,
      visitedFromStart,
      visitedFromEnd,
      forwardPaths,
      backwardPaths,
    )
    if (result.length > 0) {
      return result
    }

    // Expand the BFS from the end side
    result = await processQueue(
      backwardQueue,
      visitedFromEnd,
      visitedFromStart,
      backwardPaths,
      forwardPaths,
    )
    if (result.length > 0) {
      return result
    }
  }

  return []
}

// const getPathsBetweenPages = async (
//   fromPage: Page,
//   toPage: Page,
// ): Promise<number[][]> => {
//   const paths: number[][] = []

//   const depthFirstSearch = async (
//     currentPage: Page,
//     currentPath: number[],
//   ): Promise<void> => {
//     currentPath.push(currentPage.id)
//     if (currentPage.id === toPage.id) {
//       paths.push([...currentPath])
//     } else {
//       for (const link of currentPage.internalLinks) {
//         const isAlreadyVisited = currentPath.includes(link.id)
//         if (!isAlreadyVisited) {
//           await link.load("internalLinks")
//           await depthFirstSearch(link, currentPath)
//         }
//       }
//     }
//     currentPath.pop()
//   }

//   await depthFirstSearch(fromPage, [])
//   return paths
// }

export default class get_shortest_paths {
  public async handle(context: HttpContext): Promise<
    | {
        __response: ExceptionMessage
        __status: 404
      }
    | {
        __response: ExceptionMessage
        __status: 500
      }
    | {
        __response: get_shortest_paths_response
        __status: 200
      }
  > {
    const payload = await context.request.validateUsing(
      get_shortest_paths_validator,
    )

    const fromPage = await Page.findOrFail(payload.fromPageId)
    await fromPage.load("internalLinks")

    const toPage = await Page.findOrFail(payload.toPageId)
    await toPage.load("internalLinks")

    const isDepth0 = fromPage.id === toPage.id
    if (isDepth0) {
      return context.response.ok({
        pages: {
          [fromPage.id]: {
            id: fromPage.id,
            title: fromPage.title,
          },
        },
        paths: [],
      })
    }

    const isDepth1 = fromPage.internalLinks.some((internalLink) => {
      return internalLink.id === toPage.id
    })
    if (isDepth1) {
      return context.response.ok({
        pages: {
          [fromPage.id]: {
            id: fromPage.id,
            title: fromPage.title,
          },
          [toPage.id]: {
            id: toPage.id,
            title: toPage.title,
          },
        },
        paths: [fromPage.id, toPage.id],
      })
    }

    const paths = await getPathsBetweenPages(fromPage, toPage)

    return context.response.ok({
      pages: {
        [fromPage.id]: {
          id: fromPage.id,
          title: fromPage.title,
        },
        [toPage.id]: {
          id: toPage.id,
          title: toPage.title,
        },
      },
      paths,
    })

    // return context.response.internalServerError({
    //   message: "Shortest paths can't be determined.",
    // })
  }
}

router
  .get("/wikipedia/shortest-paths", [get_shortest_paths])
  .use(throttle)
  .openapi({
    description:
      "Find the shortest paths between two Wikipedia pages, using only internal links.",
    tags: ["wikipedia"],
  })
